'----------------------------------------------------------------------------------------
'    Name:  Modbus_Master_Functions.inc
'    Type:  TIGER-BASIC(tm) Source Code (Modbus on RS485-bus)
' Release:  V 1.0
' Purpose:  ModbusProtocol (Master) on RS485
'           (for information about the Modbus protocol please visit www.modbus.org)
'
'  (C) - Copyright Wilke Technology, P.O.Box 1727, D-52018 Aachen, Germany
'----------------------------------------------------------------------------------------
'
' Thank you for using BASIC Tigers in your products. If you have questions, ideas
' or special needs, please contact your next distributor or the Tiger support team
' and visit our web site:
'
'	Wilke Technology GmbH
'	The Tiger Support Team
'	P.O.Box 1727,       D-52018 Aachen, Germany
'	Krefelder Str. 147, D-52070 Aachen, Germany
'
'	email: support@wilke-technology.com  (english)
'	email: support@wilke.de              (german)
'	Phone: +49 (241) 918 900    Mo to Fr, 7:00 to 16:00 (GMT)
'	Fax:   +49 (241) 918 9068
'
' New information, new drivers and free downloads see:
'
'	www.wilke-technology.com	     (english)
'	www.wilke.de                         (german)
'
' Sincerely,
'
'	Your Tiger Support Team
'
'
'--------------------------------------------------------------------------------------
' Supported Mobbus-Master-Functions:        ' Function names:
' 0x01 - Read Coils                         ' ReadCoils
' 0x02 - Read Discrete Inputs               ' ReadDisInputs
' 0x03 - Read Hold Reg                      ' ReadHoldReg
' 0x04 - Read Input Reg                     ' ReadInputReg
' 0x05 - Write Single Coil                  ' WriteSingleCoil
' 0x06 - Write Single Reg                   ' WriteSingleReg
' 0x07 - Read Exception Reg                 ' ReadExceptionReg
' 0x0B - Get Com Event Counter              ' GetComEventCounter
' 0x0C - Get Com Event Log                  ' GetCommEventLog
' 0x0F - Write Multiple Coils               ' WriteMultipleCoils
' 0x10 - Write Multiple Reg                 ' WriteMultipleReg
' 0x11 - Report Slave Id                    ' ReportSlaveId
' 0x16 - Mask Write Register                ' MaskWriteReg
' 0x17 - Read Write Multiple Regs           ' ReadWriteMultipleRegs
' 0x2B+0x0E - Read Device Identification    ' ReadDeviceIdentification
'--------------------------------------------------------------------------------------
#ifndef Modbus_Protocol_Functions
#define Modbus_Protocol_Functions

#include Modbus_Defs.inc
#include Modbus_CalcCRC_i.inc
#include Modbus_Conf.inc
#include DEFINE_A.INC
#include ufunc4.inc

'' these functions are not implemented yet:
'' 08h Diagnostics
'' 14h ReadFileRecord
'' 15h WriteFileRecord
'' 02Bh/00Dh CANopen General Reference Request and Response PDU


''*************************************************************
''**** 0x01 Read Coils
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - svpReadData$ (string): contains the read data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadCoils( byte bpDeviceId; word wpStartAddress; word wpQuantity; var String svpReadData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode
    byte blByteCount

    bvpReturn = NO_ERROR

    if wpQuantity<0000h or wpQuantity > 07D0h then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif

	
    '' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_COILS)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(svpReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(svpReadData$) < 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(svpReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_COILS then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(svpReadData$,MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount + 3 <> len(svpReadData$) then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpReadData$ = mid$(svpReadData$,3,blByteCount)
    return

End


''*************************************************************
''**** 0x02 Read Discrete Inputs
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - svpReadData$ (string): contains the read data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadDisInputs( byte bpDeviceId; word wpStartAddress; word wpQuantity; var String svpReadData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode

    bvpReturn = NO_ERROR

    if wpQuantity<0000h or wpQuantity > 07D0h then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DIS_INPUTS)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(svpReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(svpReadData$) < 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(svpReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_DIS_INPUTS then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(svpReadData$,MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount + 3 <> len(svpReadData$) then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpReadData$ = mid$(svpReadData$,3,blByteCount)
    return

End



''*************************************************************
''**** 0x03 Read Hold Reg
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - svpReadData$ (string): contains the read data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadHoldReg( byte bpDeviceId; word wpStartAddress; word wpQuantity; var String svpReadData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode
    byte blByteCount

    bvpReturn = NO_ERROR

    if wpQuantity<0000h or wpQuantity > 07Dh then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_HOLD_REG)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(svpReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(svpReadData$) < 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(svpReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_HOLD_REG then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(svpReadData$,MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount+3 <> len(svpReadData$) then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpReadData$ = mid$(svpReadData$,3,blByteCount)
    return
End



''*************************************************************
''**** 0x04 Read Input Reg
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - svpReadData$ (string): contains the read data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadInputReg( byte bpDeviceId; word wpStartAddress; word wpQuantity; var String svpReadData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode
    byte blByteCount

    bpReturn = 0

    if wpQuantity<0000h or wpQuantity > 07Dh then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_INPUT_REG)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(svpReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(svpReadData$) < 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(svpReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_INPUT_REG then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(svpReadData$,MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount+3 <> len(svpReadData$) then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpReadData$ = mid$(svpReadData$,3,blByteCount)
    return

End


''*************************************************************
''**** 0x05 Write Single Coil
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpAddress (word): address for writing
''****   - wpValue (word): value to write
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub WriteSingleCoil( byte bpDeviceId; word wpAddress; word wpValue; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)

    bvpReturn = NO_ERROR

    if wpValue <> 0000h and wpValue <> 0FF00h then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpAddress>0FFFFh) or (wpAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_SINGLE_COIL)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpValue)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if slRetMsg$ <> slMBMsg$ then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    return

End



''*************************************************************
''**** 0x06 Write Single Reg
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpAddress (word): address for writing
''****   - wpValue (word): value to write
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub WriteSingleReg( byte bpDeviceId; word wpAddress; word wpValue; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)

    bvpReturn = NO_ERROR

    if wpValue < 0000h or wpValue > 0FFFFh then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpAddress>0FFFFh) or (wpAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_SINGLE_REG)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpValue)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if slRetMsg$ <> slMBMsg$ then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    return

End


''*************************************************************
''**** 0x07 Read Exception Reg
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - bvpERegData (byte): contains the read exception register's data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadExceptionReg( byte bpDeviceId; var byte bvpERegData; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slReadData$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode

    bvpReturn = NO_ERROR

	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_EXCEPTION_REG)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slReadData$) <> 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(slReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_EXCEPTION_REG then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    bvpERegData = nfroms(slReadData$,2, SZE_OF_VAR_BYTE)
    return

End


''*************************************************************
''**** 0x0B Get Com Event Counter
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wvpComEvCount (word): read amount of communication events
''****   - wvpStatus (word): current device status
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub GetComEventCounter( byte bpDeviceId; var word wvpComEvCount; var word wvpStatus; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slReadData$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode

    bvpReturn = NO_ERROR

	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_GET_COM_EVENT_COUNTER)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slReadData$) <> 6 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(slReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_GET_COM_EVENT_COUNTER then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    wvpStatus = nfroms(slReadData$, MB_STAT_POS_TRANS, -SZE_OF_VAR_WORD)
    wvpComEvCount = nfroms(slReadData$, MB_EVCT_POS_TRANS, -SZE_OF_VAR_WORD)
    return

End


''*************************************************************
''**** 0x0C Get Com Event Log
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wvpComEvCount (word): read amount of communication events
''****   - wvpStatus (word): current device status
''****   - wvpMessageCount (word): read quantity of messages processed by the device
''****   - svpEvents$ (string): read event log of the device
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub GetCommEventLog( byte bpDeviceId; var word wvpComEvCount; var word wvpStatus; var word wvpMessageCount; var String svpEvents$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slReadData$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode
    byte blByteCount

    bvpReturn = NO_ERROR

	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_GET_COM_EVENT_LOG)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slReadData$) < 9 then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    blRecFuncCode = nfroms(slReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_GET_COM_EVENT_LOG then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(slReadData$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount < 6 then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    wvpStatus = nfroms(slReadData$, MB_STAT_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD)
    wvpComEvCount = nfroms(slReadData$, MB_EVCT_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD)
    wvpMessageCount = nfroms(slReadData$, MB_MSGC_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD)

    if blByteCount > 6 then
        svpEvents$ = mid$(slReadData$,MB_EVTS_POS_TRANS,blByteCount-6)
    else
        svpEvents$ = ""
    endif

    return

End



''*************************************************************
''**** 0x0F Write Multiple Coils
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - spValues$ (string): data to write
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub WriteMultipleCoils( byte bpDeviceId; word wpStartAddress; word wpQuantity; String spValues$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)

    bvpReturn = NO_ERROR

    if wpQuantity<0000h or wpQuantity > 07B0h then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif

	
	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_MULT_COILS)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)
    slMBMsg$ = ntos$(slMBMsg$, MB_VALS_POS_TRANS, SZE_OF_VAR_BYTE, len(spValues$))

    slMBMsg$ = slMBMsg$ + spValues$

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if slRetMsg$ <> mid$(slMBMsg$,0,6) then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    return

End



''*************************************************************
''**** 0x10 Write Multiple Reg
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpStartAddress (word): start address for reading
''****   - wpQuantity (word): quantity of values to read
''****   - spValues$ (string): data to write
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub WriteMultipleReg( byte bpDeviceId; word wpStartAddress; word wpQuantity; String spValues$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)

    bvpReturn = NO_ERROR

    if wpQuantity<0000h or wpQuantity > 07Bh then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpStartAddress>0FFFFh) or (wpStartAddress+wpQuantity>0FFFFh) or &
        (wpStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif

    if (mod(len(spValues$), 2)) = 1 then
        bvpReturn = QUANTITY_ERROR
        return
    endif
	
	
	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_MULT_REG)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpQuantity)
    slMBMsg$ = ntos$(slMBMsg$, MB_VALS_POS_TRANS, SZE_OF_VAR_BYTE, len(spValues$))

    slMBMsg$ = slMBMsg$ + spValues$

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if slRetMsg$ <> mid$(slMBMsg$,0,6) then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    return

End


''*************************************************************
''**** 0x11 Report Slave Id
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - svpSlaveId$ (string): read slave identification
''****   - bvpRunIndStatus (byte): run indicator status of the device
''****   - svpAddData$ (string): additional device information
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReportSlaveId( byte bpDeviceId; var String svpSlaveId$; var byte bvpRunIndStatus; var String svpAddData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slReadData$(OUT_GOING_STR_MAX_SZE)
    byte blRecFuncCode
    byte blByteCount

    bvpReturn = NO_ERROR

	
	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_REPORT_SLAVE_ID)

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slReadData$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slReadData$) < 4 then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    blRecFuncCode = nfroms(slReadData$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_REPORT_SLAVE_ID then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(slReadData$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    svpSlaveId$ = mid$(slReadData$, MB_DVID_POS_TRANS, blByteCount)
    bvpRunIndStatus = nfroms(slReadData$, blByteCount+3, SZE_OF_VAR_BYTE)
    if bvpRunIndStatus <> 00h AND bvpRunIndStatus <> 0FFh then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpAddData$ = mid$(slReadData$, blByteCount+4, len(slReadData$)-blByteCount-4)

    return

End


''*************************************************************
''**** 0x16 Mask Write Register
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpAddress (word): address for writing
''****   - wpAndMask (word): and-mask for writing
''****   - wpOrMask (word): or-mask for wrinting
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub MaskWriteReg( byte bpDeviceId; word wpAddress; word wpAndMask; word wpOrMask; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)

    bvpReturn = NO_ERROR

	
	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_MASK_WRITE_REGISTER)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_AMSK_POS_TRANS, -SZE_OF_VAR_WORD, wpAndMask)
    slMBMsg$ = ntos$(slMBMsg$, MB_OMSK_POS_TRANS, -SZE_OF_VAR_WORD, wpOrMask)


    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if slRetMsg$ <> slMBMsg$ then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    return

End



''*************************************************************
''**** 0x17 Read Write Multiple Regs
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - wpRStartAddress (word): start address for reading
''****   - wpRQuantity (word): quantity of registers o read
''****   - wpWStartAddress (word): start address for writing
''****   - wpWQuantity (word): quantity of registers to write
''****   - spValues$ (string): values to write
''****   - svpReadData (string): read data
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadWriteMultipleRegs( byte bpDeviceId; word wpRStartAddress; word wpRQuantity; word wpWStartAddress; word wPWQuantity; String spValues$; var String svpReadData$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)
    byte blRetFuncCode

    bvpReturn = NO_ERROR

    if wpWQuantity<0000h or wpWQuantity > 079h then
        '''# ERROR ExceptionCode = 03h
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpWStartAddress>0FFFFh) or (wpWStartAddress+wpWQuantity>0FFFFh) or &
        (wpWStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


    if wpRQuantity<0000h or wpRQuantity > 07Dh then
        '''# ERROR ExceptionCode =03
        bvpReturn = QUANTITY_ERROR
        return
    endif

    if (wpRStartAddress>0FFFFh) or (wpRStartAddress+wpRQuantity>0FFFFh) or &
        (wpRStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        bvpReturn = ADDRESS_ERROR
        return
    endif


    if (mod(len(spValues$), 2)) = 1 then
        bvpReturn = QUANTITY_ERROR
        return
    endif
	
	
    '' assemble modbus message	
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_WRITE_MULT_REGS)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wpRStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS, -SZE_OF_VAR_WORD, wpRQuantity)
    slMBMsg$ = ntos$(slMBMsg$, MB_ADDR_POS_TRANS+4, -SZE_OF_VAR_WORD, wpWStartAddress)
    slMBMsg$ = ntos$(slMBMsg$, MB_RQTY_POS_TRANS+4, -SZE_OF_VAR_WORD, wpWQuantity)
    slMBMsg$ = ntos$(slMBMsg$, MB_VALS_POS_TRANS+4, SZE_OF_VAR_BYTE, len(spValues$))

    slMBMsg$ = slMBMsg$ + spValues$

    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slRetMsg$) < 3 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    blRecFuncCode = nfroms(slRetMsg$,MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
    if blRecFuncCode <> MBFC_READ_WRITE_MULT_REGS then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    blByteCount = nfroms(slRetMsg$,MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE)
    if blByteCount+3 <> len(slRetMsg$) then
        bvpReturn = CORRUPTED_PACKET_ERROR
        return
    endif

    svpReadData$ = mid$(slRetMsg$,3,blByteCount)

    return

End



''*************************************************************
''**** 0x02B+0x0E Read Device Identification
''****
''**** function values:
''****   - bpDeviceId (byte): address of the target device
''****   - bpReadDevIdCode (byte): device-id-code of the read information
''****   - bpObjectId (byte): object id of the read information
''****   - bvpConformityLevel (byte): the objects conformitylevel
''****   - bvpMoreFollows (byte): more objects to read?
''****   - bvpNextObjectId (byte): id of the next object to read
''****   - svpObjects$ (string): read object
''****   - bvpReturn (byte): returns error codes
''*************************************************************
Sub ReadDeviceIdentification( byte bpDeviceId; byte bpReadDevIdCode; byte bpObjectId; var byte bvpConformityLevel; var byte bvpMoreFollows; var byte bvpNextObjectId; var String svpObjects$; var byte bvpReturn)
    String slMBMsg$(OUT_GOING_STR_MAX_SZE)
    String slRetMsg$(OUT_GOING_STR_MAX_SZE)
    word wlFuncCode

    bvpReturn = NO_ERROR

    bvpConformityLevel = 0
    bvpMoreFollows = 0
    bvpNextObjectId = 0
    svpObjects$ = ""

    if bpReadDevIdCode < 1h AND bpReadDevIdCode > 4h then
        bvpReturn = ADDRESS_ERROR
        return
    endif

    if bpObjectId > 06h AND bpObjectId < 80h then
        bvpReturn = ADDRESS_ERROR
        return
    endif

	
	'' assemble modbus message
    slMBMsg$ = ""
    slMBMsg$ = ntos$(slMBMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, bpDeviceId)
    slMBMsg$ = ntos$(slMBMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_IDENTIFICATION)
    slMBMsg$ = ntos$(slMBMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_MEI_TYPE)
    slMBMsg$ = ntos$(slMBMsg$, MB_DVID_POS_TRANS, SZE_OF_VAR_BYTE, bpReadDevIdCode)
    slMBMsg$ = ntos$(slMBMsg$, MB_OBID_POS_TRANS, SZE_OF_VAR_BYTE, bpObjectId)


    call SendMSG(slMBMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif



    call ReceiveAnswer(slRetMsg$,bvpReturn)
    if bvpReturn <> NO_ERROR then
        return
    endif

    if len(slRetMsg$) < 8 then
        bvpReturn = CORRUPTED_PACKET_ERROR
    endif

    wlFuncCode = nfroms(slRetMsg$, MB_FUNC_POS_TRANS, -SZE_OF_VAR_WORD)
    if wlFuncCode <> MBFC_READ_DEV_IDENTIFICATION*0100h+MBFC_READ_DEV_MEI_TYPE then
        bvpReturn = WRONG_ANSWER_ERROR
        return
    endif

    bvpConformityLevel = nfroms(slRetMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE)
    bvpMoreFollows = nfroms(slRetMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE)
    bvpNextObjectId = nfroms(slRetMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE)

    svpObjects$ = mid$(slRetMsg$,MB_OBJC_POS_TRANS,len(slRetMsg$)-7)

    return

End





''*************************************************************
''**** Send Message - Sends the MB-Message to other devices
''*************************************************************
sub SendMSG( String spMBMessage$; var byte bvpReturn)
    String slMessage$(OUT_GOING_STR_MAX_SZE)
    word wlCRC

    wlCRC = 0
    bvpReturn = NO_ERROR
    slMessage$ = spMBMessage$


    call calcCRC( slMessage$, wlCRC)
    slMessage$ = ntos$(slMessage$, len(slMessage$), SZE_OF_VAR_WORD, wlCRC)


    ''CLEAR INPUT BUFFER FIRST
    PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0

    ''SEND HERE VIA RS485
    PUT #SER, #SEC_RS485, slMessage$


    ''ON ERROR SET bvpReturn (see list for values)

end


''*************************************************************
''**** Receive Message - Trys to receive an answer message
''*************************************************************
sub ReceiveAnswer( var String svpRetMsg$; var byte bvpReturn)
    String slIncoming$(OUT_GOING_STR_MAX_SZE)
    long llTimeout
    word wlMsgCRC
    word wlCRC
    word wlIBUFill
    word wlTemp

    wlCRC = 0
    wlMsgCRC = 0
    bvpReturn = NO_ERROR
    svpRetMsg$ = ""
    slIncoming$ = ""

    llTimeout = ticks + MB_RECEIVE_TIMEOUT

    while llTimeout > ticks
        GET #SER, #SEC_RS485, #UFCI_IBU_FILL, 0, wlIBUFill
        if wlIBUFill <> 0 then
            wait_duration 4

            GET #SER, #SEC_RS485, #UFCI_IBU_FILL, 0, wlTemp
            if wlTemp = wlIBUFill then

                GET #SER, #SEC_RS485, 0, slIncoming$

                if len(slIncoming$) >= 4 then
                    wlMsgCRC = nfroms(slIncoming$, len(slIncoming$)-2, SZE_OF_VAR_WORD)
                    svpRetMsg$ = left$(slIncoming$, len(slIncoming$)-2)
                    call calcCRC(svpRetMsg$,wlCRC)

                    if wlCRC <> wlMsgCRC then
                        PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0
                        bvpReturn = CRC_ERROR
                        return
                    endif
                    return
                endif
            endif
        else
            wait_duration 10
        endif
    endwhile

    PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0
    bvpReturn = TIMEOUT_ERROR
    return

end


''*************************************************************
''**** Modbus Init
''*************************************************************
Sub InitModbus()
    call CRCinit()
    INSTALL_DEVICE #SER, "SER1B_K4.TDD",&
        BAUD_SEC_0, PARITY_SEC_0, REC_FLAG_SEC_0,&
        BAUD_SEC_1, PARITY_SEC_1, REC_FLAG_SEC_1,&
        TR_EN_MASK_SEC_0,LOG_PORT_SEC_0,TR_HILO_SEC_0,&
        TR_EN_MASK_SEC_1,LOG_PORT_SEC_1,TR_HILO_SEC_1
End


#endif

